@@ -41,25 +41,33 @@ from .function import (
     AugmentedExpectedImprovement,
     BatchMonteCarloExpectedHypervolumeImprovement,
     BatchMonteCarloExpectedImprovement,
+    BayesianActiveLearningByDisagreement,
     ExpectedConstrainedHypervolumeImprovement,
     ExpectedConstrainedImprovement,
+    ExpectedFeasibility,
     ExpectedHypervolumeImprovement,
     ExpectedImprovement,
     Fantasizer,
+    GreedyContinuousThompsonSampling,
+    IntegratedVarianceReduction,
     LocalPenalization,
     MinValueEntropySearch,
     MultipleOptimismNegativeLowerConfidenceBound,
     NegativeLowerConfidenceBound,
     NegativePredictiveMean,
+    ParallelContinuousThompsonSampling,
     PredictiveVariance,
     ProbabilityOfFeasibility,
     augmented_expected_improvement,
     batch_ehvi,
+    bayesian_active_learning_by_disagreement,
+    bichon_ranjan_criterion,
     expected_hv_improvement,
     expected_improvement,
     gibbon_quality_term,
     gibbon_repulsion_term,
     hard_local_penalizer,
+    integrated_variance_reduction,
     lower_confidence_bound,
     min_value_entropy_search,
     multiple_optimism_lower_confidence_bound,
