@@ -24,6 +24,7 @@ import tensorflow_probability as tfp
 
 from ...data import Dataset
 from ...models import ProbabilisticModel
+from ...models.interfaces import HasReparamSampler, SupportsGetObservationNoise
 from ...space import SearchSpace
 from ...types import TensorType
 from ...utils import DEFAULTS
@@ -31,6 +32,7 @@ from ..interface import (
     AcquisitionFunction,
     AcquisitionFunctionBuilder,
     AcquisitionFunctionClass,
+    ProbabilisticModelType,
     SingleModelAcquisitionBuilder,
     SingleModelVectorizedAcquisitionBuilder,
 )
@@ -123,7 +125,7 @@ class expected_improvement(AcquisitionFunctionClass):
         return (self._eta - mean) * normal.cdf(self._eta) + variance * normal.prob(self._eta)
 
 
-class AugmentedExpectedImprovement(SingleModelAcquisitionBuilder[ProbabilisticModel]):
+class AugmentedExpectedImprovement(SingleModelAcquisitionBuilder[SupportsGetObservationNoise]):
     """
     Builder for the augmented expected improvement function for optimization single-objective
     optimization problems with high levels of observation noise.
@@ -135,7 +137,7 @@ class AugmentedExpectedImprovement(SingleModelAcquisitionBuilder[ProbabilisticMo
 
     def prepare_acquisition_function(
         self,
-        model: ProbabilisticModel,
+        model: SupportsGetObservationNoise,
         dataset: Optional[Dataset] = None,
     ) -> AcquisitionFunction:
         """
@@ -146,6 +148,11 @@ class AugmentedExpectedImprovement(SingleModelAcquisitionBuilder[ProbabilisticMo
             greater than one.
         :raise tf.errors.InvalidArgumentError: If ``dataset`` is empty.
         """
+        if not isinstance(model, SupportsGetObservationNoise):
+            raise NotImplementedError(
+                f"AugmentedExpectedImprovement only works with models that support "
+                f"get_observation_noise; received {model.__repr__()}"
+            )
         tf.debugging.Assert(dataset is not None, [])
         dataset = cast(Dataset, dataset)
         tf.debugging.assert_positive(len(dataset), message="Dataset must be populated.")
@@ -156,7 +163,7 @@ class AugmentedExpectedImprovement(SingleModelAcquisitionBuilder[ProbabilisticMo
     def update_acquisition_function(
         self,
         function: AcquisitionFunction,
-        model: ProbabilisticModel,
+        model: SupportsGetObservationNoise,
         dataset: Optional[Dataset] = None,
     ) -> AcquisitionFunction:
         """
@@ -175,7 +182,7 @@ class AugmentedExpectedImprovement(SingleModelAcquisitionBuilder[ProbabilisticMo
 
 
 class augmented_expected_improvement(AcquisitionFunctionClass):
-    def __init__(self, model: ProbabilisticModel, eta: TensorType):
+    def __init__(self, model: SupportsGetObservationNoise, eta: TensorType):
         r"""
         Return the Augmented Expected Improvement (AEI) acquisition function for single-objective
         global optimization under homoscedastic observation noise.
@@ -198,16 +205,7 @@ class augmented_expected_improvement(AcquisitionFunctionClass):
         """
         self._model = model
         self._eta = tf.Variable(eta)
-
-        try:
-            self._noise_variance = tf.Variable(model.get_observation_noise())
-        except NotImplementedError:
-            raise ValueError(
-                """
-                Augmented expected improvement only currently supports homoscedastic gpflow models
-                with a likelihood.variance attribute.
-                """
-            )
+        self._noise_variance = tf.Variable(model.get_observation_noise())
 
     def update(self, eta: TensorType) -> None:
         """Update the acquisition function with a new eta value and noise variance."""
@@ -346,8 +344,6 @@ class ProbabilityOfFeasibility(SingleModelAcquisitionBuilder[ProbabilisticModel]
         """
         tf.debugging.assert_scalar(threshold)
 
-        super().__init__()
-
         self._threshold = threshold
 
     def __repr__(self) -> str:
@@ -423,7 +419,7 @@ def probability_of_feasibility(
     return acquisition
 
 
-class ExpectedConstrainedImprovement(AcquisitionFunctionBuilder[ProbabilisticModel]):
+class ExpectedConstrainedImprovement(AcquisitionFunctionBuilder[ProbabilisticModelType]):
     """
     Builder for the *expected constrained improvement* acquisition function defined in
     :cite:`gardner14`. The acquisition function computes the expected improvement from the best
@@ -434,7 +430,7 @@ class ExpectedConstrainedImprovement(AcquisitionFunctionBuilder[ProbabilisticMod
     def __init__(
         self,
         objective_tag: str,
-        constraint_builder: AcquisitionFunctionBuilder[ProbabilisticModel],
+        constraint_builder: AcquisitionFunctionBuilder[ProbabilisticModelType],
         min_feasibility_probability: float | TensorType = 0.5,
     ):
         """
@@ -471,7 +467,7 @@ class ExpectedConstrainedImprovement(AcquisitionFunctionBuilder[ProbabilisticMod
 
     def prepare_acquisition_function(
         self,
-        models: Mapping[str, ProbabilisticModel],
+        models: Mapping[str, ProbabilisticModelType],
         datasets: Optional[Mapping[str, Dataset]] = None,
     ) -> AcquisitionFunction:
         """
@@ -520,7 +516,7 @@ class ExpectedConstrainedImprovement(AcquisitionFunctionBuilder[ProbabilisticMod
     def update_acquisition_function(
         self,
         function: AcquisitionFunction,
-        models: Mapping[str, ProbabilisticModel],
+        models: Mapping[str, ProbabilisticModelType],
         datasets: Optional[Mapping[str, Dataset]] = None,
     ) -> AcquisitionFunction:
         """
@@ -568,7 +564,7 @@ class ExpectedConstrainedImprovement(AcquisitionFunctionBuilder[ProbabilisticMod
         return self._constrained_improvement_fn
 
     def _update_expected_improvement_fn(
-        self, objective_model: ProbabilisticModel, feasible_mean: TensorType
+        self, objective_model: ProbabilisticModelType, feasible_mean: TensorType
     ) -> None:
         """
         Set or update the unconstrained expected improvement function.
@@ -585,7 +581,7 @@ class ExpectedConstrainedImprovement(AcquisitionFunctionBuilder[ProbabilisticMod
             self._expected_improvement_fn.update(eta)  # type: ignore
 
 
-class BatchMonteCarloExpectedImprovement(SingleModelAcquisitionBuilder[ProbabilisticModel]):
+class BatchMonteCarloExpectedImprovement(SingleModelAcquisitionBuilder[HasReparamSampler]):
     """
     Expected improvement for batches of points (or :math:`q`-EI), approximated using Monte Carlo
     estimation with the reparametrization trick. See :cite:`Ginsbourger2010` for details.
@@ -606,8 +602,6 @@ class BatchMonteCarloExpectedImprovement(SingleModelAcquisitionBuilder[Probabili
         tf.debugging.assert_positive(sample_size)
         tf.debugging.assert_greater_equal(jitter, 0.0)
 
-        super().__init__()
-
         self._sample_size = sample_size
         self._jitter = jitter
 
@@ -617,7 +611,7 @@ class BatchMonteCarloExpectedImprovement(SingleModelAcquisitionBuilder[Probabili
 
     def prepare_acquisition_function(
         self,
-        model: ProbabilisticModel,
+        model: HasReparamSampler,
         dataset: Optional[Dataset] = None,
     ) -> AcquisitionFunction:
         """
@@ -643,7 +637,7 @@ class BatchMonteCarloExpectedImprovement(SingleModelAcquisitionBuilder[Probabili
     def update_acquisition_function(
         self,
         function: AcquisitionFunction,
-        model: ProbabilisticModel,
+        model: HasReparamSampler,
         dataset: Optional[Dataset] = None,
     ) -> AcquisitionFunction:
         """
@@ -662,7 +656,7 @@ class BatchMonteCarloExpectedImprovement(SingleModelAcquisitionBuilder[Probabili
 
 
 class batch_monte_carlo_expected_improvement(AcquisitionFunctionClass):
-    def __init__(self, sample_size: int, model: ProbabilisticModel, eta: TensorType, jitter: float):
+    def __init__(self, sample_size: int, model: HasReparamSampler, eta: TensorType, jitter: float):
         """
         :param sample_size: The number of Monte-Carlo samples.
         :param model: The model of the objective function.
@@ -676,16 +670,14 @@ class batch_monte_carlo_expected_improvement(AcquisitionFunctionClass):
         """
         self._sample_size = sample_size
 
-        try:
-            sampler = model.reparam_sampler(self._sample_size)
-        except (NotImplementedError):
+        if not isinstance(model, HasReparamSampler):
             raise ValueError(
-                """
-                The batch Monte-Carlo expected improvement acquisition function
-                only supports models that implement a reparam_sampler method.
-                """
+                f"The batch Monte-Carlo expected improvement acquisition function only supports "
+                f"models that implement a reparam_sampler method; received {model.__repr__()}"
             )
 
+        sampler = model.reparam_sampler(self._sample_size)
+
         self._sampler = sampler
         self._eta = tf.Variable(eta)
         self._jitter = jitter
