@@ -23,23 +23,33 @@ from __future__ import annotations
 import copy
 from typing import Dict, Generic, Mapping, TypeVar, cast, overload
 
+import numpy as np
+import tensorflow as tf
+
 from .acquisition.rule import AcquisitionRule, EfficientGlobalOptimization
 from .bayesian_optimizer import OptimizationResult, Record
 from .data import Dataset
-from .models import ModelSpec, create_model
+from .logging import get_step_number, get_tensorboard_writer
+from .models import ModelSpec, TrainableProbabilisticModel, create_model
+from .models.config import ModelConfigType
 from .observer import OBJECTIVE
 from .space import SearchSpace
 from .types import State, TensorType
 from .utils import Ok, map_values
 
-S = TypeVar("S")
+StateType = TypeVar("StateType")
 """ Unbound type variable. """
 
-SP = TypeVar("SP", bound=SearchSpace)
+SearchSpaceType = TypeVar("SearchSpaceType", bound=SearchSpace)
 """ Type variable bound to :class:`SearchSpace`. """
 
+TrainableProbabilisticModelType = TypeVar(
+    "TrainableProbabilisticModelType", bound=TrainableProbabilisticModel, contravariant=True
+)
+""" Contravariant type variable bound to :class:`TrainableProbabilisticModel`. """
 
-class AskTellOptimizer(Generic[SP]):
+
+class AskTellOptimizer(Generic[SearchSpaceType]):
     """
     This class provides Ask/Tell optimization interface. It is designed for those use cases
     when control of the optimization loop by Trieste is impossible or not desirable.
@@ -49,9 +59,9 @@ class AskTellOptimizer(Generic[SP]):
     @overload
     def __init__(
         self,
-        search_space: SP,
+        search_space: SearchSpaceType,
         datasets: Mapping[str, Dataset],
-        model_specs: Mapping[str, ModelSpec],
+        model_specs: Mapping[str, TrainableProbabilisticModelType],
         *,
         fit_model: bool = True,
     ):
@@ -60,10 +70,9 @@ class AskTellOptimizer(Generic[SP]):
     @overload
     def __init__(
         self,
-        search_space: SP,
+        search_space: SearchSpaceType,
         datasets: Mapping[str, Dataset],
-        model_specs: Mapping[str, ModelSpec],
-        acquisition_rule: AcquisitionRule[TensorType, SP],
+        model_specs: Mapping[str, ModelConfigType],
         *,
         fit_model: bool = True,
     ):
@@ -72,11 +81,12 @@ class AskTellOptimizer(Generic[SP]):
     @overload
     def __init__(
         self,
-        search_space: SP,
+        search_space: SearchSpaceType,
         datasets: Mapping[str, Dataset],
-        model_specs: Mapping[str, ModelSpec],
-        acquisition_rule: AcquisitionRule[State[S | None, TensorType], SP],
-        acquisition_state: S | None = None,
+        model_specs: Mapping[str, TrainableProbabilisticModelType],
+        acquisition_rule: AcquisitionRule[
+            TensorType, SearchSpaceType, TrainableProbabilisticModelType
+        ],
         *,
         fit_model: bool = True,
     ):
@@ -85,9 +95,12 @@ class AskTellOptimizer(Generic[SP]):
     @overload
     def __init__(
         self,
-        search_space: SP,
-        datasets: Dataset,
-        model_specs: ModelSpec,
+        search_space: SearchSpaceType,
+        datasets: Mapping[str, Dataset],
+        model_specs: Mapping[str, ModelConfigType],
+        acquisition_rule: AcquisitionRule[
+            TensorType, SearchSpaceType, TrainableProbabilisticModelType
+        ],
         *,
         fit_model: bool = True,
     ):
@@ -96,10 +109,13 @@ class AskTellOptimizer(Generic[SP]):
     @overload
     def __init__(
         self,
-        search_space: SP,
-        datasets: Dataset,
-        model_specs: ModelSpec,
-        acquisition_rule: AcquisitionRule[TensorType, SP],
+        search_space: SearchSpaceType,
+        datasets: Mapping[str, Dataset],
+        model_specs: Mapping[str, TrainableProbabilisticModelType],
+        acquisition_rule: AcquisitionRule[
+            State[StateType | None, TensorType], SearchSpaceType, TrainableProbabilisticModelType
+        ],
+        acquisition_state: StateType | None,
         *,
         fit_model: bool = True,
     ):
@@ -108,11 +124,93 @@ class AskTellOptimizer(Generic[SP]):
     @overload
     def __init__(
         self,
-        search_space: SP,
+        search_space: SearchSpaceType,
+        datasets: Mapping[str, Dataset],
+        model_specs: Mapping[str, ModelConfigType],
+        acquisition_rule: AcquisitionRule[
+            State[StateType | None, TensorType], SearchSpaceType, TrainableProbabilisticModelType
+        ],
+        acquisition_state: StateType | None = None,
+        *,
+        fit_model: bool = True,
+    ):
+        ...
+
+    @overload
+    def __init__(
+        self,
+        search_space: SearchSpaceType,
         datasets: Dataset,
-        model_specs: ModelSpec,
-        acquisition_rule: AcquisitionRule[State[S | None, TensorType], SP],
-        acquisition_state: S | None = None,
+        model_specs: TrainableProbabilisticModelType,
+        *,
+        fit_model: bool = True,
+    ):
+        ...
+
+    @overload
+    def __init__(
+        self,
+        search_space: SearchSpaceType,
+        datasets: Dataset,
+        model_specs: ModelConfigType,
+        *,
+        fit_model: bool = True,
+    ):
+        ...
+
+    @overload
+    def __init__(
+        self,
+        search_space: SearchSpaceType,
+        datasets: Dataset,
+        model_specs: TrainableProbabilisticModelType,
+        acquisition_rule: AcquisitionRule[
+            TensorType, SearchSpaceType, TrainableProbabilisticModelType
+        ],
+        *,
+        fit_model: bool = True,
+    ):
+        ...
+
+    @overload
+    def __init__(
+        self,
+        search_space: SearchSpaceType,
+        datasets: Dataset,
+        model_specs: TrainableProbabilisticModelType,
+        acquisition_rule: AcquisitionRule[
+            TensorType, SearchSpaceType, TrainableProbabilisticModelType
+        ],
+        *,
+        fit_model: bool = True,
+    ):
+        ...
+
+    @overload
+    def __init__(
+        self,
+        search_space: SearchSpaceType,
+        datasets: Dataset,
+        model_specs: TrainableProbabilisticModelType,
+        acquisition_rule: AcquisitionRule[
+            State[StateType | None, TensorType], SearchSpaceType, TrainableProbabilisticModelType
+        ],
+        acquisition_state: StateType | None = None,
+        *,
+        fit_model: bool = True,
+    ):
+        ...
+
+    @overload
+    def __init__(
+        self,
+        search_space: SearchSpaceType,
+        datasets: Dataset,
+        model_specs: ModelConfigType,
+        acquisition_rule: AcquisitionRule[
+            State[StateType | None, TensorType], SearchSpaceType, TrainableProbabilisticModelType
+        ],
+        acquisition_state: StateType | None,
         *,
         fit_model: bool = True,
     ):
@@ -120,12 +218,16 @@ class AskTellOptimizer(Generic[SP]):
 
     def __init__(
         self,
-        search_space: SP,
+        search_space: SearchSpaceType,
         datasets: Mapping[str, Dataset] | Dataset,
         model_specs: Mapping[str, ModelSpec] | ModelSpec,
-        acquisition_rule: AcquisitionRule[TensorType | State[S | None, TensorType], SP]
+        acquisition_rule: AcquisitionRule[
+            TensorType | State[StateType | None, TensorType],
+            SearchSpaceType,
+            TrainableProbabilisticModelType,
+        ]
         | None = None,
-        acquisition_state: S | None = None,
+        acquisition_state: StateType | None = None,
         *,
         fit_model: bool = True,
     ):
@@ -168,7 +270,9 @@ class AskTellOptimizer(Generic[SP]):
             )
 
         self._datasets = datasets
-        self._models = map_values(create_model, model_specs)
+        self._models = cast(
+            Dict[str, TrainableProbabilisticModelType], map_values(create_model, model_specs)
+        )
 
         if acquisition_rule is None:
             if self._datasets.keys() != {OBJECTIVE}:
@@ -178,7 +282,8 @@ class AskTellOptimizer(Generic[SP]):
                 )
 
             self._acquisition_rule = cast(
-                AcquisitionRule[TensorType, SP], EfficientGlobalOptimization()
+                AcquisitionRule[TensorType, SearchSpaceType, TrainableProbabilisticModelType],
+                EfficientGlobalOptimization(),
             )
         else:
             self._acquisition_rule = acquisition_rule
@@ -198,11 +303,15 @@ class AskTellOptimizer(Generic[SP]):
     @classmethod
     def from_record(
         cls,
-        record: Record[S],
-        search_space: SP,
-        acquisition_rule: AcquisitionRule[TensorType | State[S | None, TensorType], SP]
+        record: Record[StateType],
+        search_space: SearchSpaceType,
+        acquisition_rule: AcquisitionRule[
+            TensorType | State[StateType | None, TensorType],
+            SearchSpaceType,
+            TrainableProbabilisticModelType,
+        ]
         | None = None,
-    ) -> AskTellOptimizer[SP]:
+    ) -> AskTellOptimizer[SearchSpaceType]:
         """Creates new :class:`~AskTellOptimizer` instance from provided optimization state.
         Model training isn't triggered upon creation of the instance.
 
@@ -222,13 +331,13 @@ class AskTellOptimizer(Generic[SP]):
         return cls(
             search_space,
             record.datasets,
-            record.models,
+            cast(Mapping[str, TrainableProbabilisticModelType], record.models),
             acquisition_rule=acquisition_rule,  # type: ignore
             acquisition_state=record.acquisition_state,
             fit_model=False,
         )
 
-    def to_record(self) -> Record[S]:
+    def to_record(self) -> Record[StateType]:
         """Collects the current state of the optimization, which includes datasets,
         models and acquisition state (if applicable).
 
@@ -240,10 +349,10 @@ class AskTellOptimizer(Generic[SP]):
             datasets=self._datasets, models=models_copy, acquisition_state=acquisition_state_copy
         )
 
-    def to_result(self) -> OptimizationResult[S]:
+    def to_result(self) -> OptimizationResult[StateType]:
         """Converts current state of the optimization
         into a :class:`~trieste.data.OptimizationResult` object."""
-        record: Record[S] = self.to_record()
+        record: Record[StateType] = self.to_record()
         return OptimizationResult(Ok(record), [])
 
     def ask(self) -> TensorType:
@@ -291,3 +400,21 @@ class AskTellOptimizer(Generic[SP]):
             dataset = self._datasets[tag]
             model.update(dataset)
             model.optimize(dataset)
+
+        summary_writer = get_tensorboard_writer()
+        step_number = get_step_number()
+        if summary_writer:
+            with summary_writer.as_default():
+                for tag in self._datasets:
+                    with tf.name_scope(f"{tag}.model"):
+                        self._models[tag].log()
+                    tf.summary.scalar(
+                        f"{tag}.observation.best_overall",
+                        np.min(self._datasets[tag].observations),
+                        step=step_number,
+                    )
+                    tf.summary.scalar(
+                        f"{tag}.observation.best_new",
+                        np.min(new_data[tag].observations),
+                        step=step_number,
+                    )
